__author__ = "Jackson Eshbaugh"
__version__ = "05/23/2024"
"""
This file contains code pertaining to narrative generation for the Reddit TTS Bot (including scraping from Reddit).
"""

import os
import random
import time
from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.firefox.service import Service as FirefoxService
from selenium.webdriver.firefox.options import Options as FirefoxOptions
from selenium.webdriver.firefox.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement

from . import tts


class Narrative:
    """
    This class represents a narrative object, which will eventually be converted to an audio file.

    Attributes:
    text: The text to be spoken
    title: The title of the story
    word_count: The word count of the story
    """

    def __init__(self, text: str, title: str):
        """
        This function initializes a Narrative object.
        :param text: The text to be spoken
        :param title: The title of the story
        """
        self.text = text
        self.title = title
        self.word_count = len(text.split())

    def title(self) -> str:
        """
        This function returns the title of the story.
        :return: The title of the story
        """
        return self.title

    def word_count(self) -> int:
        """
        This function returns the word count of the story.
        :return: The word count of the story
        """
        return self.word_count

    def text(self) -> str:
        """
        This function returns the text of the story.
        :return: The text of the story
        """
        return self.text

    def to_audio(self, file_path: str, max_words_per_part: int = 140) -> list[str]:
        """
        This function converts the text of the story to an audio file.
        :param file_path: The path to save the .wav file (parent directory)
        :return: The name(s) of the .wav file(s)
        """

        if self.word_count > max_words_per_part:
            # Split the story into parts
            parts = []
            words = self.text.split()
            part = ""
            for word in words:
                if len(part.split()) < max_words_per_part:
                    part += word + " "
                else:
                    parts.append(part)
                    part = word + " "
            parts.append(part)

            # Save each part as an audio file
            for i, part in enumerate(parts):
                tts.tts(part, file_path, self.title + " Part " + str(i + 1))
            return [self.title + ' Part ' + str(i + 1) + '.wav' for i in range(len(parts))]
        else:
            tts.tts(self.text, file_path, self.title)
            return [self.title + '.wav']


# Scrape narratives from Reddit

locations: list = ["r/tifu", "r/entitledparents", "r/AmItheAsshole"]

SCROLL_PAUSE_TIME: int = 3


def scrape_narratives(num_narratives: int, min_word_count: int = 100,
                      locations: list = locations) -> list:
    """
    This function scrapes narratives from Reddit.
    :param num_narratives: The number of narratives to scrape
    :param min_word_count: The minimum word count for a post
    :param locations: The locations (subreddits) to scrape from
    :return: A list of Narrative objects
    """

    # Create a file to hold the names of all the narratives ever scraped
    if not os.path.exists("narrative_names.txt"):
        with open("narrative_names.txt", "w") as file:
            file.write("")

    narratives = []

    # Set up the Selenium WebDriver for Firefox
    options = FirefoxOptions()
    options.add_argument("--start-maximized")
    driver: WebDriver = webdriver.Firefox(service=FirefoxService(), options=options)

    try:
        for i in range(num_narratives):
            # Choose a random location to pull stories from
            location = random.choice(locations)

            print("Scraping from " + location)

            # Open the URL
            driver.get(f"https://www.reddit.com/{location}/top/?t=all")

            # Get the articles
            posts: list[WebElement] = driver.find_elements(By.TAG_NAME, 'article')

            # Choose a random post
            post: WebElement = random.choice(posts)

            while post.get_attribute('aria-label') in open("narrative_names.txt").read():
                # Scroll the page a bit to load more content
                driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
                time.sleep(SCROLL_PAUSE_TIME)
                posts = driver.find_elements(By.TAG_NAME, 'article')
                post = random.choice(posts)

            # Get the title (stored in the aria-label attribute of each post)
            title: str = post.get_attribute("aria-label").replace(":", "-").replace("/", "-")

            # Write the title to the file to keep track of which posts have been scraped
            with open("narrative_names.txt", "a") as file:
                file.write(title + "\n")

            # Get the content: go to the post's page and scrape the content
            post_url: str = post.find_element(By.TAG_NAME, "a").get_attribute("href")
            driver.get(post_url)
            time.sleep(3)  # Wait for the page to load
            post_soup: BeautifulSoup = BeautifulSoup(driver.page_source, "html.parser")
            content_elem = post_soup.select_one('.md.text-14')
            content: str = content_elem.text.strip() if content_elem else ""
            content = content.replace('\n', ' ')

            # Check if the post meets the minimum word count requirement
            if len(content.split()) < min_word_count:
                i -= 1
                continue

            # Add the title to the beginning of the content so that it is spoken
            content = title + ". " + content

            # Create a Narrative object
            narrative = Narrative(content, title)
            narratives.append(narrative)

    finally:
        driver.quit()

    return narratives



__author__ = "Jackson Eshbaugh"
__version__ = "05/24/2024"
"""
This file contains code pertaining to video generation for the Reddit TTS Bot.
"""

import os
import random
import pandas as pd

import cv2
import moviepy.editor as mp
import whisper
from moviepy.video.VideoClip import TextClip
from moviepy.audio.io.AudioFileClip import AudioFileClip
from moviepy.video.VideoClip import VideoClip
from moviepy.video.compositing.CompositeVideoClip import CompositeVideoClip
from moviepy.video.io.VideoFileClip import VideoFileClip
from moviepy.video.tools.subtitles import SubtitlesClip

from narrative import Narrative


def get_random_video() -> str | None:
    """
    This function returns a random video from the videos directory.
    :return: The name of the video file or None if no videos are available
    """

    if not os.path.exists("videos"):
        raise FileNotFoundError("The videos directory does not exist.")

    videos = os.listdir("videos")
    if not videos:
        raise FileNotFoundError("No videos found in the videos directory.")

    return random.choice(videos)


def subtitle_video(video_file: str, video_directory: str) -> None:
    """
    Uses OpenAI's Whisper model to generate subtitles for a video.

    This function is modified from Gradient AI's Whisper-AutoCaption repository: https://github.com/gradient-ai/Whisper-AutoCaption/blob/main/app.py
    :param video_file: The video file to subtitle (including the file extension)
    :param video_directory: The directory containing the video file
    :return: None
    """
    my_clip = mp.VideoFileClip(os.getcwd() + os.sep + '.temp_video' + os.sep + video_file)

    # Create a temporary directory for the audio file
    os.makedirs(os.getcwd() + os.sep + '.temp_subtitle', exist_ok=True)

    if len(os.listdir(os.getcwd() + os.sep + '.temp_subtitle' + os.sep)) == 0:
        my_clip.audio.write_audiofile(os.getcwd() + os.sep + '.temp_subtitle' + os.sep + 'audio.mp3',
                                      codec="libmp3lame")

    # Instantiate whisper model using model_type variable
    model = whisper.load_model('medium')

    # Get text from speech for subtitles from audio file
    result = model.transcribe(os.getcwd() + os.sep + '.temp_subtitle' + os.sep + 'audio.mp3', task='translate')

    # create Subtitle dataframe, and save it
    dict1 = {'start': [], 'end': [], 'text': []}
    for i in result['segments']:
        dict1['start'].append(int(i['start']))
        dict1['end'].append(int(i['end']))
        dict1['text'].append(i['text'])
    df = pd.DataFrame.from_dict(dict1)
    vidcap = cv2.VideoCapture(os.getcwd() + os.sep + '.temp_video' + os.sep + video_file)
    success, image = vidcap.read()
    height = image.shape[0]
    width = image.shape[1]

    # Instantiate MoviePy subtitle generator with TextClip, subtitles, and SubtitlesClip
    generator = lambda txt: TextClip(txt, font='P052-Bold', fontsize=width / 20, stroke_width=1, color='white',
                                     stroke_color='black', size=(width, height * .25), method='caption')
    subs = tuple(zip(tuple(zip(df['start'].values, df['end'].values)), df['text'].values))
    subtitles = SubtitlesClip(subs, generator)

    # Create final video directory if it doesn't exist
    os.makedirs(video_directory, exist_ok=True)

    video: VideoFileClip = VideoFileClip(os.getcwd() + os.sep + '.temp_video' + os.sep + video_file)
    final: CompositeVideoClip = CompositeVideoClip([video, subtitles.set_pos(('center', 'center'))])
    final.write_videofile(video_directory + os.sep + video_file, fps=video.fps, remove_temp=True, codec="libx264",
                          audio_codec="aac")

    # remove the temporary audio files and folder
    os.system("rm -r .temp_subtitle")


def add_video_to_narrative(narrative: Narrative, video_directory: str = "output") -> bool:
    """
    This function adds a randomly selected video to a WAV audio file.
    :param audio_file_name: The name of the audio file (excluding the file extension)
    :param audio_file_path: The path to the audio file (parent directory)
    :return: None
    """

    audio_files = narrative.to_audio(file_path=os.getcwd() + os.sep + ".temp_audio")

    os.makedirs('.temp_video', exist_ok=True)

    for audio_file in audio_files:
        audio_file = audio_file.replace(":", "-").replace("/", "-")
        random_video: str | None = get_random_video()

        if random_video is None:
            raise FileNotFoundError("No videos found in the \"videos\" directory.")

        video: VideoFileClip(VideoClip) = mp.VideoFileClip("videos" + os.sep + random_video)
        audio: AudioFileClip = mp.AudioFileClip(os.getcwd() + os.sep + ".temp_audio" + os.sep + audio_file)

        while video.duration < audio.duration:
            random_video = get_random_video()
            video = mp.VideoFileClip("videos" + os.sep + random_video)

        # Select a random time interval within the video equal to the length of the audio file
        start_time = random.uniform(0, video.duration - audio.duration)
        video = video.subclip(start_time, start_time + audio.duration)

        # Overwrite any existing audio in the video with the new audio
        video = video.set_audio(audio)

        # Resize the video to the desired height while maintaining aspect ratio
        video = video.resize(height=1920)

        # Crop the video to 1080x1920 (9:16 aspect ratio)
        # The crop will be centered, so x1 will be (video.size[0]-1080)/2
        video = video.crop(x1=(video.size[0] - 1080) / 2, width=1080)

        # Write the new video to the output directory
        output = os.getcwd() + os.sep + '.temp_video' + os.sep + audio_file[:len(audio_file) - 4] + ".mp4"
        os.makedirs(os.path.dirname(output), exist_ok=True)

        video.write_videofile(output, codec="libx264", audio_codec="aac")
        subtitle_video(video_file=audio_file[:len(audio_file) - 4] + ".mp4", video_directory=video_directory)

    # delete the temporary audio files and folder
    os.system("rm -r .temp_audio")
    os.system("rm -r .temp_video")

    return True



__author__ = "Jackson Eshbaugh"
__version__ = "05/22/2024"
"""
This file contains code pertaining to text-to-speech (TTS) functionality for the Reddit TTS Bot.
"""

import os

from voicebox import SimpleVoicebox
from voicebox.tts import gTTS
from voicebox.sinks import Distributor, WaveFile


def tts(text: str, file_path: str, file_name: str) -> None:
    """
    This function takes a string of text and saves it as a .wav file.
    :param text: The text to be spoken
    :param file_path: The path to save the .wav file (parent directory)
    :param file_name: The name of the .wav file (excluding the file extension)
    :return: None
    """

    # Create directory if it doesn't exist
    if not os.path.exists(file_path):
        os.makedirs(file_path)

    sink = Distributor([
        WaveFile(file_path + os.sep + file_name + ".wav")
    ])
    voicebox = SimpleVoicebox(tts=gTTS(), sink=sink)
    voicebox.say(text)



__author__ = "Jackson Eshbaugh"
__version__ = "05/24/2024"

import os

from setuptools import setup, find_packages

setup(
    name='reddit_tts_bot',
    version='1.0.5',
    packages=find_packages(),
    description='A module that facilitates the creation of short form content from Reddit posts.',
    long_description=open(os.path.dirname(__file__) + os.sep + 'README.md').read(),
    long_description_content_type='text/markdown',
    author='Jackson Eshbaugh',
    author_email='jacksoneshbaugh@gmail.com',
    url='https://github.com/jacksoneshbaugh/reddit-tts-bot',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'attrs>=23.2.0',
        'beautifulsoup4>=4.12.3',
        'bs4>=0.0.2',
        'cachetools>=5.3.3',
        'certifi>=2024.2.2',
        'cffi>=1.16.0',
        'charset-normalizer>=3.3.2',
        'click>=8.1.7',
        'decorator>=4.4.2',
        'filelock>=3.14.0',
        'fsspec>=2024.5.0',
        'gTTS>=2.5.1',
        'h11>=0.14.0',
        'idna>=3.7',
        'imageio>=2.34.1',
        'imageio-ffmpeg>=0.4.9',
        'Jinja2>=3.1.4',
        'joblib>=1.4.2',
        'llvmlite>=0.42.0',
        'MarkupSafe>=2.1.5',
        'more-itertools>=10.2.0',
        'moviepy>=1.0.3',
        'mpmath>=1.3.0',
        'networkx>=3.3',
        'nltk>=3.8.1',
        'numba>=0.59.1',
        'numpy>=1.26.4',
        'openai-whisper>=20231117',
        'opencv-python>=4.9.0.80',
        'outcome>=1.3.0.post0',
        'pandas>=2.2.2',
        'pedalboard>=0.9.6',
        'pillow>=10.3.0',
        'proglog>=0.1.10',
        'pycparser>=2.22',
        'pydub>=0.25.1',
        'PySocks>=1.7.1',
        'python-dateutil>=2.9.0.post0',
        'pytz>=2024.1',
        'regex>=2024.5.15',
        'requests>=2.32.2',
        'scipy>=1.13.0',
        'selenium>=4.21.0',
        'six>=1.16.0',
        'sniffio>=1.3.1',
        'sortedcontainers>=2.4.0',
        'sounddevice>=0.4.6',
        'soupsieve>=2.5',
        'sympy>=1.12',
        'tiktoken>=0.7.0',
        'torch>=2.3.0',
        'tqdm>=4.66.4',
        'trio>=0.25.1',
        'trio-websocket>=0.11.1',
        'typing_extensions>=4.11.0',
        'tzdata>=2024.1',
        'urllib3>=2.2.1',
        'voicebox-tts>=0.0.7',
        'wsproto>=1.2.0'
    ],
)
